
from __future__ import print_function, division

import sys
from collections import Counter

class SimplePerceptron(object):
    """Basic perceptron implementation."""

    def __init__(self, n_iter):
        self.n_iter = n_iter

    def fit(self, X, Y):
        self.w = Counter()
        for i in range(self.n_iter):
          for x, y in zip(X, Y):
            score = sum(self.w[word] for word in x)
            if score <= 0 and y == 'pos':
                for word in x:
                    self.w[word] += 1
            elif score >= 0 and y == 'neg':
                for word in x:
                    self.w[word] -= 1
            #print(self.w)
          #sys.exit(0)
        
    def predict_one(self, x):
        score = sum(self.w[word] for word in x)
        if score > 0:
            return 'pos'
        else:
            return 'neg'
    
    def predict(self, X):        
        return [ self.predict_one(x) for x in X ]

    def print_strongest_features(self, n=10):
        print('Strongest positive:')
        for f in sorted(self.w, key=self.w.get, reverse=True)[:n]:
            print(' ', f, self.w[f])
        print('Strongest negative:')
        for f in sorted(self.w, key=self.w.get, reverse=False)[:n]:
            print(' ', f, self.w[f])


### Reimplementation of the perceptron using NumPy/SciPy vectors ###

import numpy

class LinearClassifier(object):
    """Base class for linear classifiers, i.e. classifiers with a linear
    prediction function. See slide 45 in the lecture."""

    def find_classes(self, Y):
        """Find the set of output classes in a training set.

        The attributes positive_class and negative_class will be assigned
        after calling this method. An exception is raised if the number of
        classes in the training set is not equal to 2."""

        classes = sorted(set(Y))
        if len(classes) != 2:
            raise Exception("this does not seem to be a 2-class problem")
        self.positive_class = classes[0]
        self.negative_class = classes[1]

    def predict(self, X):        
        """Apply the linear scoring function and outputs self.positive_class
        for instances where the scoring function is positive, otherwise
        self.negative_class."""
        
        # To speed up, we apply the scoring function to all the instances
        # at the same time.
        scores = X.dot(self.w)
        
        # Create the output array.
        # At the positions where the score is positive, this will contain
        # self.positive class, otherwise self.negative_class.
        out = numpy.select([scores>=0.0, scores<0.0], [self.positive_class, 
                                                       self.negative_class])
        return out


class NewPerceptron(LinearClassifier):
    """A perceptron implementation, see slide 47 in the lecture."""

    def __init__(self, n_iter=10):
        self.n_iter = n_iter

    def fit(self, X, Y):
        Y = list(Y)
        self.find_classes(Y)

        # The feature vectors returned by DictVectorizer/CountVectorizer
        # are sparse vectors of the type scipy.sparse.csc_matrix. We convert
        # them to dense vectors of the type numpy.array.
        X = X.toarray()

        # The shape attribute holds the dimensions of the feature matrix. 
        # This is a tuple where
        # X.shape[0] = number of instances, 
        # X.shape[1] = number of features
        n_features = X.shape[1]

        # Initialize the weight vector to all zero
        self.w = numpy.zeros( n_features )

        for i in range(self.n_iter):
          for x, y in zip(X, Y):
            
            # the dot product is now used instead of
            # the dictionary-based scoring we saw before:
            # score = sum(self.w[word] for word in x)
            score = self.w.dot(x)
            
            if score <= 0 and y == self.positive_class:
                self.w += x
            elif score >= 0 and y == self.negative_class:
                self.w -= x

def sign(y, pos):
    if y == pos:
        return 1.0
    else:
        return -1.0

class NewPerceptron2(LinearClassifier):
    """Reformulation of the perceptron where we encode the output classes as
    +1 and -1, see slide 48 in the lecture."""

    def __init__(self, n_iter=10):
        self.n_iter = n_iter

    def fit(self, X, Y):
        Y = list(Y)
        self.find_classes(Y)

        # convert all output values to +1 or -1
        Yn = [sign(y, self.positive_class) for y in Y]

        X = X.toarray()
        n_features = X.shape[1]
        self.w = numpy.zeros( n_features )
        
        for i in range(self.n_iter):            
            for x, y in zip(X, Yn):
                score = self.w.dot(x)
                if y*score <= 0:
                    self.w += y*x


# Two helper functions for processing sparse and dense vectors.
# I haven't been able to do this efficiently in a more "civilised" manner: 
# these functions rely on the internal details of scipy.sparse.csr_matrix.
def add_sparse_to_dense(x, w, xw):
    w[x.indices] += xw*x.data
def sparse_dense_dot(x, w):
    return numpy.dot(w[x.indices], x.data)

class SparsePerceptron(LinearClassifier):
    """A perceptron implementation using sparse feature vectors, see 
    slide 50 in the lecture."""

    def __init__(self, n_iter=10):
        self.n_iter = n_iter

    def fit(self, X, Y):
        Y = list(Y)
        self.find_classes(Y)

        Yn = [sign(y, self.positive_class) for y in Y]
        n_features = X.shape[1]
        self.w = numpy.zeros( n_features )
        X = list(X) 
        for i in range(self.n_iter):            
            for x, y in zip(X, Yn):
                score = sparse_dense_dot(x, self.w)
                if y*score <= 0:
                    add_sparse_to_dense(x, self.w, y)
